use crate::{expression::Expression, interpreter::Interpreter, statement::Statement, token::Token};
use std::collections::HashMap;
use thiserror::Error;

#[derive(Error, Debug)]
pub enum ResolverError {
    #[error("line {0}: can not read local variable in its own initializer.")]
    ReadVarInOwnInitializer(usize),
    #[error("line {0}: already a variable with this name in this scope.")]
    NameRedefinition(usize),
    #[error("line {0}: can not return from top-level code.")]
    TopLevelReturn(usize),
    #[error("line {0}: can not use 'this' outside of a class.")]
    ThisOutsideClass(usize),
    #[error("line {0}: can not return a value from an initializer.")]
    ReturnFromInitializer(usize),
    #[error("line {0}: a class can not inherit from itself.")]
    SelfInheritance(usize),
    #[error("line {0}: can not use 'super' outside of a class.")]
    SuperOutsideClass(usize),
    #[error("line {0}: can not use 'super' in a class with no superclass.")]
    SuperWithoutSuperclass(usize),
}

#[derive(Debug, Clone, Copy, PartialEq, Eq)]
enum FunctionType {
    None,
    Function,
    Initializer,
    Method,
}

#[derive(Debug, Clone, Copy, PartialEq, Eq)]
enum ClassType {
    None,
    Class,
    Subclass,
}

/// Resolve variable references and perform static analysis before interpretation.
#[derive(Debug)]
pub struct Resolver<'a> {
    /// Stack of scopes, each mapping variable names to their initialized state.
    scopes: Vec<HashMap<String, bool>>,
    /// Reference to the interpreter to resolve variable depths.
    interpreter: &'a mut Interpreter,
    /// Track the type of the current function being resolved (regular, method, initializer, ...).
    current_fun: FunctionType,
    /// Tracks the type of the current class being resolved (regular, subclass, ...).
    current_class: ClassType,
}

impl<'a> Resolver<'a> {
    /// Create a new resolver with a reference to the interpreter
    pub fn new(interpreter: &'a mut Interpreter) -> Self {
        Self {
            scopes: Vec::default(),
            interpreter,
            current_fun: FunctionType::None,
            current_class: ClassType::None,
        }
    }

    /// Resolve all statements in a program
    pub fn resolve(&mut self, statements: &[Statement]) -> Result<(), ResolverError> {
        for stm in statements {
            self.resolve_statement(stm)?;
        }
        Ok(())
    }

    /// Resolve a single statement.
    fn resolve_statement(&mut self, statement: &Statement) -> Result<(), ResolverError> {
        match statement {
            Statement::Block(statements) => self.resolve_block(statements),
            Statement::Print(expression) => self.resolve_expression(expression),
            Statement::Expression(expression) => self.resolve_expression(expression),
            Statement::Var { name, initializer } => {
                self.resolve_var(name, (**initializer).as_ref())
            }
            Statement::If {
                condition,
                then_branch,
                else_branch,
            } => self.resolve_if_stmt(
                condition,
                then_branch,
                else_branch.as_ref().map(|b| b.as_ref()),
            ),
            Statement::While { condition, body } => self.resolve_while_stmt(condition, body),
            Statement::Function { name, params, body } => {
                self.resolve_function_stmt(name, params, body)
            }
            Statement::Return { keyword, value } => self.resolve_return_stmt(keyword, value),
            Statement::Class {
                name,
                superclass,
                methods,
            } => self.resolve_class_stmt(name, superclass, methods),
            Statement::For {
                initializer,
                condition,
                increment,
                body,
            } => self.resolve_for_statement(initializer, condition, increment, body),
        }
    }

    /// Resolve an expression.
    fn resolve_expression(&mut self, expr: &Expression) -> Result<(), ResolverError> {
        match expr {
            Expression::Assign { name, value } => self.resolve_assign_expr(expr, name, value),
            Expression::Binary {
                left,
                operator: _,
                right,
            } => self.resolve_binary_expr(left, right),
            Expression::Call {
                callee,
                paren: _,
                args,
            } => self.resolve_call_expr(callee, args),
            Expression::Grouping { expression } => self.resolve_expression(expression),
            Expression::Literal { value: _ } => Ok(()),
            Expression::Logical {
                left,
                operator: _,
                right,
            } => self.resolve_binary_expr(left, right),
            Expression::Unary { operator: _, right } => self.resolve_expression(right),
            Expression::Variable { name } => self.resolve_var_expr(expr, name),
            Expression::Get { object, name: _ } => self.resolve_expression(object),
            Expression::Set {
                object,
                name: _,
                value,
            } => self.resolve_set_expression(object, value),
            Expression::This { keyword } => {
                if self.current_class == ClassType::None {
                    return Err(ResolverError::ThisOutsideClass(keyword.line));
                }

                self.resolve_local(expr, keyword)
            }
            Expression::Super { keyword, method: _ } => {
                self.resolve_super_expression(keyword, expr)
            }
        }
    }

    /// Create a new scope for variable declarations.
    fn begin_scope(&mut self) {
        self.scopes.push(HashMap::new());
    }

    /// Remove the current scope.
    fn end_scope(&mut self) {
        self.scopes.pop();
    }

    /// Declare a variable in the current scope and mark it as not initialized yet (false).
    ///
    /// Error if the variable is already declared in this scope.
    fn declare(&mut self, name: &Token) -> Result<(), ResolverError> {
        if let Some(scope) = self.scopes.last_mut() {
            if scope.contains_key(&name.lexeme) {
                return Err(ResolverError::NameRedefinition(name.line));
            }

            scope.insert(name.lexeme.clone(), false);
        }

        Ok(())
    }

    /// Mark a variable as initialized (true) in the current scope.
    fn define(&mut self, name: &Token) {
        if let Some(scope) = self.scopes.last_mut() {
            scope.insert(name.lexeme.clone(), true);
        }
    }

    /// Resolve a block statement by creating a new scope.
    /// Resolve all statements in the block, then remove the scope.
    fn resolve_block(&mut self, statements: &[Statement]) -> Result<(), ResolverError> {
        self.begin_scope();
        self.resolve(statements)?;
        self.end_scope();

        Ok(())
    }

    /// Resolve a class declaration with methods and optional inheritance.
    /// Handle special cases for 'this', 'super', and initializers.
    fn resolve_class_stmt(
        &mut self,
        name: &Token,
        superclass: &Option<Expression>,
        methods: &[Statement],
    ) -> Result<(), ResolverError> {
        let enclosing_class = self.current_class;
        self.current_class = ClassType::Class;

        self.declare(name)?;
        self.define(name);

        if let Some(
            superclass @ Expression::Variable {
                name: superclass_name,
            },
        ) = superclass
        {
            if superclass_name.lexeme == name.lexeme {
                return Err(ResolverError::SelfInheritance(superclass_name.line));
            }

            self.current_class = ClassType::Subclass;
            self.resolve_expression(superclass)?;

            self.begin_scope();
            if let Some(scope) = self.scopes.last_mut() {
                scope.insert("super".to_string(), true);
            }
        }

        self.begin_scope();

        if let Some(scope) = self.scopes.last_mut() {
            scope.insert("this".to_string(), true);
        }

        for method in methods {
            if let Statement::Function { name, params, body } = method {
                let declaration = if name.lexeme == "init" {
                    FunctionType::Initializer
                } else {
                    FunctionType::Method
                };
                self.resolve_function(params, body, declaration)?;
            }
        }

        self.end_scope();

        if superclass.is_some() {
            self.end_scope();
        }

        self.current_class = enclosing_class;

        Ok(())
    }

    /// Resolve a return statement, checking for invalid returns (top-level code and value returns from initializers).
    fn resolve_return_stmt(
        &mut self,
        keyword: &Token,
        value: &Option<Expression>,
    ) -> Result<(), ResolverError> {
        if self.current_fun == FunctionType::None {
            return Err(ResolverError::TopLevelReturn(keyword.line));
        }
        if let Some(value) = value {
            if self.current_fun == FunctionType::Initializer {
                return Err(ResolverError::ReturnFromInitializer(keyword.line));
            }
            self.resolve_expression(value)
        } else {
            Ok(())
        }
    }

    /// Declare the variable, resolve its initializer if any, then define it.
    fn resolve_var(
        &mut self,
        name: &Token,
        initializer: Option<&Expression>,
    ) -> Result<(), ResolverError> {
        self.declare(name)?;

        if let Some(init_expr_ref) = initializer {
            self.resolve_expression(init_expr_ref)?;
        }

        self.define(name);
        Ok(())
    }

    /// Resolve a variable reference expression.
    /// Check that the variable is not being used in its own initializer.
    fn resolve_var_expr(&mut self, expr: &Expression, name: &Token) -> Result<(), ResolverError> {
        if let Some(scope) = self.scopes.last() {
            if !scope.get(&name.lexeme).unwrap_or(&true) {
                return Err(ResolverError::ReadVarInOwnInitializer(name.line));
            }
        }
        self.resolve_local(expr, name)?;
        Ok(())
    }

    /// Resolve an assignment expression.
    /// Resolve the value being assigned, then resolve the variable being assigned to.
    fn resolve_assign_expr(
        &mut self,
        expr: &Expression,
        name: &Token,
        val: &Expression,
    ) -> Result<(), ResolverError> {
        self.resolve_expression(val)?;
        self.resolve_local(expr, name)?;
        Ok(())
    }

    /// Resolve a binary expression by resolving both operands.
    fn resolve_binary_expr(
        &mut self,
        left: &Expression,
        right: &Expression,
    ) -> Result<(), ResolverError> {
        self.resolve_expression(left)?;
        self.resolve_expression(right)?;
        Ok(())
    }

    /// Resolve a function declaration.
    /// Declare and define the function name, then resolve the function body.
    fn resolve_function_stmt(
        &mut self,
        name: &Token,
        params: &[Token],
        body: &[Statement],
    ) -> Result<(), ResolverError> {
        self.declare(name)?;
        self.define(name);
        self.resolve_function(params, body, FunctionType::Function)?;

        Ok(())
    }

    /// Resolve a function call expression.
    /// Resolve the callee and all arguments.
    fn resolve_call_expr(
        &mut self,
        callee: &Expression,
        args: &[Expression],
    ) -> Result<(), ResolverError> {
        self.resolve_expression(callee)?;

        for arg in args {
            self.resolve_expression(arg)?;
        }

        Ok(())
    }

    /// Resolve a property assignment expression.
    /// Resolve both the object and the value being assigned.
    fn resolve_set_expression(
        &mut self,
        object: &Expression,
        value: &Expression,
    ) -> Result<(), ResolverError> {
        self.resolve_expression(value)?;
        self.resolve_expression(object)?;

        Ok(())
    }

    /// Resolve a super expression, checking for invalid uses (outside of classes, classes with no superclass).
    fn resolve_super_expression(
        &mut self,
        keyword: &Token,
        expression: &Expression,
    ) -> Result<(), ResolverError> {
        if self.current_class == ClassType::None {
            return Err(ResolverError::SuperOutsideClass(keyword.line));
        } else if self.current_class != ClassType::Subclass {
            return Err(ResolverError::SuperWithoutSuperclass(keyword.line));
        }

        self.resolve_local(expression, keyword)
    }

    /// Resolve an if statement by resolving the condition and both branches.
    fn resolve_if_stmt(
        &mut self,
        condition: &Expression,
        then_branch: &Statement,
        else_branch: Option<&Statement>,
    ) -> Result<(), ResolverError> {
        self.resolve_expression(condition)?;
        self.resolve_statement(then_branch)?;

        if let Some(stmt_ref) = else_branch {
            self.resolve_statement(stmt_ref)?;
        }

        Ok(())
    }

    /// Resolve a while statement by resolving the condition and body.
    fn resolve_while_stmt(
        &mut self,
        condition: &Expression,
        body: &Statement,
    ) -> Result<(), ResolverError> {
        self.resolve_expression(condition)?;
        self.resolve_statement(body)?;

        Ok(())
    }

    /// Resolve a for statement.
    /// This creates a new scope for the entire for construct.
    fn resolve_for_statement(
        &mut self,
        initializer: &Option<Box<Statement>>,
        condition: &Option<Expression>,
        increment: &Option<Expression>,
        body: &Statement,
    ) -> Result<(), ResolverError> {
        self.begin_scope();

        if let Some(init_stmt) = initializer {
            self.resolve_statement(init_stmt)?;
        }

        if let Some(cond_expr) = condition {
            self.resolve_expression(cond_expr)?;
        }

        self.resolve_statement(body)?;

        if let Some(inc_expr) = increment {
            self.resolve_expression(inc_expr)?;
        }

        self.end_scope();
        Ok(())
    }

    /// Resolve a function definition with parameters and body.
    /// Create a new scope for the function body and declare all parameters.
    fn resolve_function(
        &mut self,
        params: &[Token],
        body: &[Statement],
        fun_type: FunctionType,
    ) -> Result<(), ResolverError> {
        let enclosing_fun = self.current_fun;
        self.current_fun = fun_type;

        self.begin_scope();

        for param in params {
            self.declare(param)?;
            self.define(param);
        }
        self.resolve(body)?;

        self.end_scope();

        self.current_fun = enclosing_fun;

        Ok(())
    }

    /// Resolve a local variable reference by finding its scope.
    /// Tell the interpreter the number of scopes between the reference and definition.
    fn resolve_local(
        &mut self,
        expression: &Expression,
        name: &Token,
    ) -> Result<(), ResolverError> {
        for (idx, scope) in self.scopes.iter().rev().enumerate() {
            if scope.contains_key(&name.lexeme) {
                self.interpreter.resolve(expression.clone(), idx);
                return Ok(());
            }
        }

        Ok(())
    }
}
