//! Scan source code to create tokens out of it.

use crate::{
    token::{Token, TokenType::Eof},
    tokenizer::{
        comment::Comment, identifier::Identifier, interface::Tokenizer, lookahead::Lookahead,
        newline::Newline, number::Number, single_char::SingleChar, string::String,
        whitespace::Whitespace,
    },
};
use lazy_static::lazy_static;
use tracing::error;

lazy_static! {
    /// Tokenizers to use in scanning. They are tried in the exact order in which they appear in
    /// the list.
    static ref TOKENIZERS: Vec<Box<dyn Tokenizer>> = vec![
        Box::new(SingleChar),
        Box::new(Whitespace),
        Box::new(Newline),
        Box::new(Lookahead),
        Box::new(Comment),
        Box::new(String),
        Box::new(Number),
        Box::new(Identifier),
    ];
}

/// Take source code as input and return a list of tokens representing it.
pub fn tokenize(source: &str) -> Vec<Token> {
    let mut tokens: Vec<Token> = Vec::new();
    let mut source_chars = source.char_indices().peekable();

    let mut line = 1;
    while let Some(c) = source_chars.next() {
        // careful, a tokenizer run can move the iterator but I have not found a more ergonomic variant yet
        match TOKENIZERS
            .iter()
            .find_map(|x| x.run(c, &mut source_chars, source, line))
        {
            Some((line_advance, token)) => {
                // I do not like handling it this way, but it suffices for now
                line += line_advance;

                if let Some(token) = token {
                    tokens.push(token);
                }
            }
            None => {
                error!("unexpected character: {}", c.1)
            }
        }
    }
    // Eof is always the last token
    tokens.push(Token {
        token_type: Eof,
        lexeme: "".to_string(),
        literal: None,
        line,
    });

    tokens
}

#[cfg(test)]
mod tests {
    use ordered_float::OrderedFloat;

    use crate::token::Literal;
    use crate::token::Token;
    use crate::token::TokenType::*;

    use super::tokenize;

    const FIBONACCI: &str = r#"
      // first 21 elements in the Fibonacci sequence
      var a = 0;
      var temp;

      for (var b = 1; a < 10000; b = temp + b) {
        print a;
        temp = a;
        a = b;
      }
    "#;

    #[test]
    fn floating_points() {
        assert_eq!(
            vec![Token {
                token_type: Number,
                lexeme: "0".into(),
                literal: Some(Literal::Number(OrderedFloat(0.0))),
                line: 1
            },],
            vec![Token {
                token_type: Number,
                lexeme: "0".into(),
                literal: Some(Literal::Number(OrderedFloat(0.0))),
                line: 1
            },]
        );
    }

    #[test]
    fn multiline_string() {
        let input = r#""Hello,
 world
!""#;
        let tokens = tokenize(input);

        assert_eq!(
            tokens,
            vec![
                Token {
                    token_type: String,
                    lexeme: "\"Hello,\n world\n!\"".into(),
                    literal: Some(Literal::String("Hello,\n world\n!".into())),
                    line: 1
                },
                Token {
                    token_type: Eof,
                    lexeme: "".into(),
                    literal: None,
                    line: 3
                },
            ]
        );
    }

    #[test]
    fn fibonacci_tokens() {
        let tokens = tokenize(FIBONACCI);
        assert_eq!(
            tokens,
            vec![
                Token {
                    token_type: Var,
                    lexeme: "var".into(),
                    literal: None,
                    line: 3
                },
                Token {
                    token_type: Identifier,
                    lexeme: "a".into(),
                    literal: None,
                    line: 3
                },
                Token {
                    token_type: Equal,
                    lexeme: "=".into(),
                    literal: None,
                    line: 3
                },
                Token {
                    token_type: Number,
                    lexeme: "0".into(),
                    literal: Some(Literal::Number(OrderedFloat(0.0))),
                    line: 3
                },
                Token {
                    token_type: Semicolon,
                    lexeme: ";".into(),
                    literal: None,
                    line: 3
                },
                Token {
                    token_type: Var,
                    lexeme: "var".into(),
                    literal: None,
                    line: 4
                },
                Token {
                    token_type: Identifier,
                    lexeme: "temp".into(),
                    literal: None,
                    line: 4
                },
                Token {
                    token_type: Semicolon,
                    lexeme: ";".into(),
                    literal: None,
                    line: 4
                },
                Token {
                    token_type: For,
                    lexeme: "for".into(),
                    literal: None,
                    line: 6
                },
                Token {
                    token_type: LeftParen,
                    lexeme: "(".into(),
                    literal: None,
                    line: 6
                },
                Token {
                    token_type: Var,
                    lexeme: "var".into(),
                    literal: None,
                    line: 6
                },
                Token {
                    token_type: Identifier,
                    lexeme: "b".into(),
                    literal: None,
                    line: 6
                },
                Token {
                    token_type: Equal,
                    lexeme: "=".into(),
                    literal: None,
                    line: 6
                },
                Token {
                    token_type: Number,
                    lexeme: "1".into(),
                    literal: Some(Literal::Number(OrderedFloat(1.0))),
                    line: 6
                },
                Token {
                    token_type: Semicolon,
                    lexeme: ";".into(),
                    literal: None,
                    line: 6
                },
                Token {
                    token_type: Identifier,
                    lexeme: "a".into(),
                    literal: None,
                    line: 6
                },
                Token {
                    token_type: Less,
                    lexeme: "<".into(),
                    literal: None,
                    line: 6
                },
                Token {
                    token_type: Number,
                    lexeme: "10000".into(),
                    literal: Some(Literal::Number(OrderedFloat(10000.0))),
                    line: 6
                },
                Token {
                    token_type: Semicolon,
                    lexeme: ";".into(),
                    literal: None,
                    line: 6
                },
                Token {
                    token_type: Identifier,
                    lexeme: "b".into(),
                    literal: None,
                    line: 6
                },
                Token {
                    token_type: Equal,
                    lexeme: "=".into(),
                    literal: None,
                    line: 6
                },
                Token {
                    token_type: Identifier,
                    lexeme: "temp".into(),
                    literal: None,
                    line: 6
                },
                Token {
                    token_type: Plus,
                    lexeme: "+".into(),
                    literal: None,
                    line: 6
                },
                Token {
                    token_type: Identifier,
                    lexeme: "b".into(),
                    literal: None,
                    line: 6
                },
                Token {
                    token_type: RightParen,
                    lexeme: ")".into(),
                    literal: None,
                    line: 6
                },
                Token {
                    token_type: LeftBrace,
                    lexeme: "{".into(),
                    literal: None,
                    line: 6
                },
                Token {
                    token_type: Print,
                    lexeme: "print".into(),
                    literal: None,
                    line: 7
                },
                Token {
                    token_type: Identifier,
                    lexeme: "a".into(),
                    literal: None,
                    line: 7
                },
                Token {
                    token_type: Semicolon,
                    lexeme: ";".into(),
                    literal: None,
                    line: 7
                },
                Token {
                    token_type: Identifier,
                    lexeme: "temp".into(),
                    literal: None,
                    line: 8
                },
                Token {
                    token_type: Equal,
                    lexeme: "=".into(),
                    literal: None,
                    line: 8
                },
                Token {
                    token_type: Identifier,
                    lexeme: "a".into(),
                    literal: None,
                    line: 8
                },
                Token {
                    token_type: Semicolon,
                    lexeme: ";".into(),
                    literal: None,
                    line: 8
                },
                Token {
                    token_type: Identifier,
                    lexeme: "a".into(),
                    literal: None,
                    line: 9
                },
                Token {
                    token_type: Equal,
                    lexeme: "=".into(),
                    literal: None,
                    line: 9
                },
                Token {
                    token_type: Identifier,
                    lexeme: "b".into(),
                    literal: None,
                    line: 9
                },
                Token {
                    token_type: Semicolon,
                    lexeme: ";".into(),
                    literal: None,
                    line: 9
                },
                Token {
                    token_type: RightBrace,
                    lexeme: "}".into(),
                    literal: None,
                    line: 10
                },
                Token {
                    token_type: Eof,
                    lexeme: "".into(),
                    literal: None,
                    line: 11
                },
            ]
        );
    }
}
